from __future__ import annotations
from pathlib import Path
import numpy as np
from typing import Any, Optional, Tuple, Union, List, Iterator, Literal, ClassVar, Annotated, get_args
from PIL import Image
import atexit, shutil, tempfile, weakref, os, time, sys, copy
from pydantic import BaseModel, ConfigDict, Field, PrivateAttr, BeforeValidator

from shinier.utils import uint8_plus
from shinier.base import ImageListType, InformativeBaseModel
from shinier.color.Converter import rgb2gray, RGB2GRAY_WEIGHTS
from shinier.Options import ACCEPTED_IMAGE_FORMATS
ACCEPTED_IMAGE_FORMATS = [f".{ext.lower()}" for ext in get_args(ACCEPTED_IMAGE_FORMATS)]


# -----------------------------------------------------------------------------
# Temp management utilities (unchanged)
# -----------------------------------------------------------------------------
_TEMP_ROOT: Optional[Path] = None
_LIVE_DIRS: set[Path] = set()


def cleanup_all_temp_dirs(max_age_hours: int = 0) -> None:
    """
    Force-remove all shinier temp roots (/tmp/shinier-<pid>) regardless of ownership.
    Optionally keep very recent ones by setting max_age_hours > 0.

    Args:
        max_age_hours (int): Minimum age (in hours) for folders to be removed.
                             0 means remove everything immediately.
    """
    tmp_root = Path(tempfile.gettempdir())
    now = time.time()
    for p in tmp_root.glob("shinier-*"):
        if not p.is_dir():
            continue
        age_hours = (now - p.stat().st_mtime) / 3600.0
        if max_age_hours == 0 or age_hours >= max_age_hours:
            shutil.rmtree(p, ignore_errors=True)


def _pid_alive(pid: int) -> bool:
    """Check if a process ID is still alive."""
    try:
        if sys.platform.startswith("win"):
            return True
        os.kill(pid, 0)
        return True
    except OSError:
        return False


def _sweep_stale_roots(max_age_hours: int = 168) -> None:
    """Remove old /tmp/shinier-<pid> roots from previous or crashed runs."""
    tmp = Path(tempfile.gettempdir())
    now = time.time()
    for p in tmp.glob("shinier-*"):
        if not p.is_dir():
            continue
        try:
            pid = int(p.name.split("-")[-1])
        except ValueError:
            continue
        age_ok = (now - p.stat().st_mtime) < (max_age_hours * 3600)
        if (not sys.platform.startswith("win") and not _pid_alive(pid)) or not age_ok:
            shutil.rmtree(p, ignore_errors=True)


def _ensure_temp_root() -> Path:
    """Create a single process-scoped temp root lazily."""
    global _TEMP_ROOT
    if _TEMP_ROOT is None:
        _sweep_stale_roots()
        _TEMP_ROOT = Path(tempfile.gettempdir()) / f"shinier-{os.getpid()}"
        _TEMP_ROOT.mkdir(parents=True, exist_ok=True)
        atexit.register(_cleanup_process_root)
    return _TEMP_ROOT


def _register_temp_dir(p: Path) -> None:
    _LIVE_DIRS.add(p)


def _unregister_temp_dir(p: Path) -> None:
    _LIVE_DIRS.discard(p)


def _cleanup_process_root() -> None:
    """Remove all registered temporary dirs and the process root at exit."""
    for p in list(_LIVE_DIRS):
        shutil.rmtree(p, ignore_errors=True)
    if _TEMP_ROOT and _TEMP_ROOT.exists():
        shutil.rmtree(_TEMP_ROOT, ignore_errors=True)


class ImageListIO(InformativeBaseModel):
    """
    Class to manage a list of images with read and write capabilities.
    Inspired by the skimage.io.ImageCollection class.

    Args:
        input_data (ImageListType):
            File pattern, list of file paths, or list of in-memory NumPy arrays.
        conserve_memory (Optional[bool]): If True (default), uses a temporary directory to store images
            and keeps only one image in memory at a time.
        as_gray (Optional[int]): Images are converted into grayscale then uint8 on load only.
            0 = No conversion
            1 = Equal weighted sum of R,G,B
            2 = Rec. ITU-R 601
            3 = Rec. ITU-R 709
            4 = Rec. ITU-R 2020
        save_dir (Optional[str]): Directory to save final images. Defaults to cwd.

    Notes:
        Validity of the collection's image attributes (i.e. dtype, n_dims, n_channels, reference_size) is only guaranteed
            after first initialization. Then, only reference_size is checked. This provides flexibility to update the
            collection with different data type and different channels (e.g. bool -> float). See more information below.
        __getitem__:
            - Height and width of the item should match reference_size.
            - Item will be automatically casted to float if collection dtype is float
            - n_channels is neither validated nor checked.
        __setitem__:
            - Height and width of the new item should match reference_size.
            - New item dtype is accepted as is and determines collection dtype.
            - n_channels is neither validated nor checked.
    """

    # --- Pydantic config ---
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        validate_assignment=True,
        extra="forbid"
    )

    # --- Static class constants ---
    DEFAULT_GRAY_MODE: ClassVar[str] = "L"
    DEFAULT_COLOR_MODE: ClassVar[str] = "RGB"
    ACCEPTED_EXTENSIONS: ClassVar[List[str]] = ACCEPTED_IMAGE_FORMATS

    # --- User-provided attributes ---
    input_data: ImageListType
    conserve_memory: bool = True
    as_gray: Literal[0, 1, 2, 3, 4] = 0
    save_dir: Optional[Path] = None

    # --- Public runtime attributes ---
    data: List[Optional[np.ndarray]] = Field(default_factory=list)
    src_paths: List[Optional[Path]] = Field(default_factory=list)
    store_paths: List[Optional[Path]] = Field(default_factory=list)
    reference_size: Optional[Tuple[int, int]] = Field(default=None)
    n_channels: Optional[int] = Field(default=None)
    n_dims: Optional[int] = Field(default=None)
    n_images: int = Field(default=0)
    dtype: Optional[np.dtype] = Field(default=None)
    drange: Optional[Tuple[Any, Any]] = Field(default=None)
    has_list_array: bool = Field(default=False)

    # --- Private runtime attributes ---
    _temp_dir: Optional[Path] = PrivateAttr(default=None)
    _finalizer: Any = PrivateAttr(default=None)
    _read_only: bool = PrivateAttr(default=False)

    # ------------------------------------------------------------------
    # Post-init constructor (replaces manual __init__)
    # ------------------------------------------------------------------
    def post_init(self, __context: Any) -> None:
        """Perform initialization and collection setup after Pydantic validation."""
        self.data = []
        self.save_dir = Path(self.save_dir or Path.cwd())
        self._initialize_collection(self.input_data)  # This also runs _initial_validation

    # ------------------------------------------------------------------
    # Core methods
    # ------------------------------------------------------------------
    def __getitem__(self, idx: int) -> np.ndarray:
        """ Access an image by index from list or store_paths. Cast to collection dtype if necessary."""
        if idx < -self.n_images or idx >= self.n_images:
            raise IndexError("Index out of range.")
        if idx < 0:
            idx = self.n_images + idx
        if self.data[idx] is None:
            if self.conserve_memory:
                # If conserve memory, keep only one image in memory
                self._reset_data()
                img = self._validate_image_attr(self._load_image(self.store_paths[idx]), attr_names=['reference_size'])

                # Cast to float if necessary
                if img.dtype != self.dtype and np.issubdtype(self.dtype, np.floating):
                    img = img.astype(self.dtype, copy=False)

                # img = self._validate_image(self._load_image(self.store_paths[idx]))

                # # Cast to collection dtype if not compatible
                # if self.dtype is not None and img.dtype != self.dtype:
                #     img = img.astype(self.dtype, copy=False)
                self.data[idx] = img
            else:
                raise ValueError(f"Data at index {idx} is None. This should not happen when conserve_memory is False.")

        return self.data[idx]

    def __setitem__(self, idx: int, new_image: np.ndarray) -> None:
        """Modify an image at a given index. Do not cast to current dtype — modify dataset dtype with new image.dtype"""
        if getattr(self, "_read_only", False):
            raise RuntimeError("This ImageListIO fork is read-only; cannot modify items.")

        if idx < 0 or idx >= self.n_images:
            raise IndexError("Index out of range.")

        # Must be the same height and width
        new_image = self._validate_image_attr(new_image, attr_names=['reference_size'])

        # Update collection dtype, n_channels and n_dims (but not enforced: attributes might thus not be accurate)
        self._update_image_attr(new_image, attr_names=['dtype', 'n_channels', 'n_dims'])

        # new_image = self._validate_image(new_image)
        # new_image = self._to_gray(new_image)
        if self.conserve_memory:
            self._reset_data()
            self._save_image(idx, new_image, save_dir=self._temp_dir)

        self.data[idx] = new_image

    def __len__(self) -> int:
        """ Get the number of images in the collection. """
        return self.n_images

    def __iter__(self) -> Iterator[np.ndarray]:
        """ Iterate over the images in the collection. """
        for idx in range(self.n_images):
            yield self[idx]

    def readonly_copy(self):
        """Produce a read-only copy of the instance."""
        # Use Pydantic's safe model duplication
        new = self.model_copy(deep=True)

        # Reset runtime-only attributes
        new._temp_dir = None
        new._finalizer = None

        # Mark as read-only by wrapping __setitem__
        new._read_only = True

        # Clear pixel arrays for memory efficiency
        new.data = [None] * self.n_images

        # No temp/staging is carried over
        new._staging_dir = None
        new._staging_paths = None
        new._staging_enabled = False

        return new

    def new_copy(self, to_list: bool = False) -> ImageListIO:
        # Construct a new instance with or without list of images.
        # This will run normal model validation and initialization.
        input_data = self.to_list() if to_list else self.input_data

        # Initialize new instance with original input_data
        new_instance = self.__class__(input_data=input_data, conserve_memory=self.conserve_memory, as_gray=self.as_gray)

        # Update new instance's image data
        if not to_list:
            for idx, im in enumerate(self):
                new_instance[idx] = im
        return new_instance

    def copy_with_image_list(self) -> ImageListIO:
        # Construct a new instance. This will run normal model validation and initialization.
        # return self.__class__(input_data=self.to_list(), conserve_memory=False)
        return self.new_copy(to_list=True)

    def to_list(self):
        """Produce a list of numpy arrays, one for each image in the collection."""
        return list(self)

    def _ensure_temp_dir(self):
        """Create this instance's temp dir lazily under the process root."""
        if self._temp_dir is None:
            root = _ensure_temp_root()
            self._temp_dir = Path(tempfile.mkdtemp(dir=root, prefix="imagelist-"))
            _register_temp_dir(self._temp_dir)
            # backstop: remove on GC
            self._finalizer = weakref.finalize(self, self._cleanup_finalizer, self._temp_dir)

    @staticmethod
    def _cleanup_finalizer(path: Path):
        try:
            shutil.rmtree(path, ignore_errors=True)
        finally:
            _unregister_temp_dir(path)

    def _cleanup_temp_dir(self) -> None:
        """Idempotent explicit cleanup."""
        if getattr(self, '_temp_dir', None) is None:
            return
        try:
            if self._temp_dir.exists():
                shutil.rmtree(self._temp_dir, ignore_errors=True)
        finally:
            _unregister_temp_dir(self._temp_dir)
            self._temp_dir = None
            if self._finalizer:
                self._finalizer()  # mark finalizer as done
                self._finalizer = None

    def _validate_image_attr(self, image: np.ndarray, attr_names: List[str], update_nonexistant: bool = True) -> np.ndarray:
        """ Validate the image and return it. """
        for attr_name in attr_names:
            value = getattr(image, attr_name, None)
            if value is None:
                if update_nonexistant:
                    self._update_image_attr(image, [attr_name])
            else:
                if attr_name == "reference_size":
                    if value != image.shape[:2]:
                        raise ValueError(f"Image size {image.shape[:2]} does not match reference size {value}.")
                elif attr_name == "n_dims":
                    if value != image.ndim:
                        raise ValueError(f"Image ndim {image.ndim} do not match collection n_dims {value}.")
                elif attr_name == "n_channels":
                    n_channels = image.shape[2] if value == 3 else 1
                    if value != n_channels:
                        raise ValueError(f"Image has {n_channels} channels which does not match collection n_channels {value}.")
                elif attr_name == "dtype":
                    if value != image.dtype:
                        raise ValueError(f"Image dtype {image.dtype} does not match collection dtype {value}.")
        return image

    def to_gray(self):
        """Public-facing version of _to_gray() that converts all images to grayscale."""
        for idx, img in enumerate(self.images):
            self.images[idx] = self._to_gray(img)

    def _to_gray(self, image: np.ndarray) -> np.ndarray:
        if image.ndim == 3:
            if self.as_gray > 0:
                image = rgb2gray(image, conversion_type=RGB2GRAY_WEIGHTS['int2key'][self.as_gray])
                image = uint8_plus(image)
        return image

    def _is_image(self, file: Path) -> bool:
        """Check if a file is an image."""
        return file.is_file() and len(file.suffix) > 1 and file.suffix.lower() in self.ACCEPTED_EXTENSIONS

    def _initial_validation(self, image: np.ndarray) -> np.ndarray:
        """Update image attribute if non-existent and make sure all images have the same attributes"""
        self._update_image_attr(image, ['reference_size', 'n_dims', 'n_channels', 'dtype'], force_update=False)
        return self._validate_image_attr(image, ['reference_size', 'n_dims', 'n_channels', 'dtype'], update_nonexistant=False)

    def _initialize_collection(self, input_data: ImageListType) -> None:
        """ Initialize the image collection from input data. """

        # Type checks
        if not (isinstance(input_data, (str, Path)) or (isinstance(input_data, list) and all(isinstance(d, (np.ndarray, str, Path)) for d in input_data))):
            raise ValueError("Input must be str|Path (glob) or list of str|Path|np.ndarray")
        if isinstance(input_data, list) and all(isinstance(d, np.ndarray) for d in input_data):
            if not all(
                    np.issubdtype(d.dtype, np.integer) or
                    np.issubdtype(d.dtype, np.floating) or
                    np.issubdtype(d.dtype, np.bool_)
                    for d in input_data):
                raise ValueError("Unsupported dtype in list of images")

        # Initialize collection
        if isinstance(input_data, (str, Path)):
            # Convert to Path if input_data is a string
            input_path = Path(input_data)

            # Handle cases with wildcards
            if "*" in str(input_path):  # Check if it's a glob pattern
                directory = input_path.parent
                pattern = input_path.name
                all_files = sorted(directory.glob(pattern))
            else:
                all_files = [input_path] if input_path.is_file() else sorted(input_path.glob("*"))
            # Filter to include only recognized image files
            self.src_paths = sorted([p for p in all_files if self._is_image(p)])
            if not self.src_paths:
                raise FileNotFoundError(f"No files found matching pattern '{input_data}'")

            # Initial store: if not conserving memory yet, the files themselves are the store
            self.store_paths = list(self.src_paths)
            self.n_images = len(self.store_paths)
        elif isinstance(input_data, list):
            self.n_images = len(input_data)
            if all(isinstance(item, np.ndarray) for item in input_data):
                # No provenance for in-memory arrays
                self.src_paths = [None] * self.n_images
                if self.conserve_memory:
                    # Write temp .npy files; those become the backing store
                    self._ensure_temp_dir()
                    self.store_paths = [self._temp_dir / f'image_{idx}.npy' for idx in range(self.n_images)]
                    for idx, im in enumerate(input_data):
                        self._save_image(idx, self._to_gray(im), save_dir=self._temp_dir)
                    self._reset_data()  # Data will not be stored in self.data when conserve_memory is True
                    self.data[0] = self._initial_validation(self._load_image(self.store_paths[0]))
                else:
                    self.has_list_array = True
                    self.data = [self._to_gray(self._initial_validation(image)) for image in input_data]
                    self._update_drange()

                    if self.store_paths.__len__() == 0:
                        self.store_paths = [None] * self.n_images
            elif all(isinstance(item, (str, Path)) for item in input_data):
                self.src_paths = [Path(item) for item in input_data]
                self.store_paths = list(self.src_paths)
            else:
                raise TypeError("input_data must be a file pattern, list of file paths, or list of NumPy arrays.")
        else:
            raise TypeError("input_data must be a file pattern, list of file paths, or list of NumPy arrays.")

        if len(self.store_paths) != self.n_images:
            self.store_paths = [None] * self.n_images

        if not self.data or all(d is None for d in self.data):
            if self.conserve_memory:
                # Only load the first image to initialize attributes
                self._reset_data()  # Data will not be stored in self.data when conserve_memory is True
                self.data[0] = self._initial_validation(self._load_image(self.store_paths[0]))
            elif not self.data and all([isinstance(fp, (str, Path)) for fp in self.store_paths]):
                # Load all images into self.data --- This could not happen:
                self.data = [self._initial_validation(self._load_image(fpath)) for fpath in self.store_paths]
            elif not all([isinstance(d, np.ndarray) for d in self.data]):
                raise ValueError('Input data should be either a list of np.ndarray or a glob pattern or a list of Path.')

        # Make sure all values have been initialized
        self._initial_validation(self.data[0])

    def _update_image_attr(self, image: np.ndarray, attr_names: List[str], force_update: bool = True) -> None:
        """Update image attributes if force_update is True or attribute is None"""
        for attr_name in attr_names:
            if attr_name == 'n_dims':
                self.n_dims = image.ndim if self.n_dims is None or force_update else self.n_dims
            elif attr_name == 'n_channels':
                n_channels = image.shape[-1] if image.ndim == 3 else 1
                self.n_channels = n_channels if self.n_channels is None or force_update else self.n_channels
            elif attr_name == 'reference_size':
                self.reference_size = image.shape[:2] if self.reference_size is None or force_update else self.reference_size
            elif attr_name == 'dtype':
                self.dtype = image.dtype if self.dtype is None or force_update else self.dtype
                self._update_drange()
            elif attr_name == 'drange':
                self._update_drange()
            elif attr_name == 'n_images':
                self.n_images = len(self.data) if self.n_images is None or force_update else self.n_images

    def _update_drange(self) -> None:
        """Update numeric dynamic range based on current self.dtype."""
        if self.dtype is None:
            self.drange = None
            return
        if np.issubdtype(self.dtype, np.bool_) or self.dtype is bool:
            self.drange = (0, 1)
        elif np.issubdtype(self.dtype, np.integer):
            info = np.iinfo(self.dtype)
            self.drange = (int(info.min), int(info.max))
        elif np.issubdtype(self.dtype, np.floating):
            # Do not assume a fixed range for floats
            pass
        else:
            self.drange = None

    def _load_image(self, image_path: Path) -> np.ndarray:
        """ Load an image from a file path. """
        image = None
        try:
            if image_path.suffix == ".npy":
                image = np.load(image_path)
                # self.dtype = image.dtype
            else:
                with Image.open(image_path) as pil_image:
                    # Load as RGB and convert to grayscale if required
                    pil_image = pil_image.convert(self.DEFAULT_COLOR_MODE)
                    image = np.array(pil_image)
                    image = self._to_gray(image)

                # self.dtype = image.dtype
            # self._update_drange()
        except IOError as e:
            raise IOError(f"Failed to load image from {image_path}: {e}")

        return image

    def _save_image(self, idx: int, image: np.ndarray, save_dir: Optional[Path] = None) -> None:
        """ Save an image to the temporary directory. """
        if getattr(self, "_read_only", False):
            raise RuntimeError("This ImageListIO fork is read-only; cannot modify items.")

        if save_dir is None and self.conserve_memory:
            self._ensure_temp_dir()
            save_dir = self._temp_dir
        else:
            save_dir = Path(save_dir or self.save_dir)
        save_dir.mkdir(parents=True, exist_ok=True)
        # save_dir = Path(save_dir or self._temp_dir or self.store_paths[idx].parent or Path.cwd())
        # save_dir.mkdir(parents=True, exist_ok=True)
        try:
            # Choose a base name:
            #   1) if we have a provenance name, keep it
            #   2) else reuse current store name
            #   3) else synthesize
            base_name = (
                    (self.src_paths[idx].name if self.src_paths[idx] is not None else None)
                    or (self.store_paths[idx].name if idx < len(self.store_paths) else None)
                    or f"image_{idx}.npy"
            )
            # base_name = self.store_paths[idx].name
            image_path = save_dir / base_name
            file_format = self.get_file_format(image_path)
            self.store_paths[idx] = image_path # Update file path
            try:
                if file_format == '.npy':
                    np.save(image_path, image.squeeze())
                else:
                    image = Image.fromarray(image.squeeze())
                    image.save(image_path, format=file_format)
            except (IOError, TypeError) as e:
                raise IOError(f"Failed to save image at index {idx} to {image_path}: {e}")
        except AttributeError as e:
            raise AttributeError(f"Failed to save image at index {idx}: {e}")

    def _reset_data(self) -> None:
        """ Reset data attribute with placeholders. """
        self.data = [None] * self.n_images

    @staticmethod
    def get_file_format(image_path: Path) -> str:
        """ Get the file format based on the file extension. """
        ext = image_path.suffix.lower()
        format_mapping = {
            '.jpg': 'JPEG', '.jpeg': 'JPEG', '.png': 'PNG',
            '.bmp': 'BMP', '.tiff': 'TIFF', '.tif': 'TIFF', '.npy': '.npy'
        }
        return format_mapping.get(ext, 'PNG')

    def final_save_all(self) -> None:
        """ Save images to save_dir. If needed (self.conserve_memory) loads images and clears up temp files. """
        for idx in range(self.n_images):
            self._save_image(idx, self[idx], save_dir=self.save_dir)

        # Clean up temporary directory
        self._cleanup_temp_dir()

    def close(self) -> None:
        self._cleanup_temp_dir()

    def __del__(self) -> None:
        """ Clean up the temporary directory upon object destruction. """
        self._cleanup_temp_dir()



from pathlib import Path
from typing import Optional, Any, List, Callable, Tuple, get_args
import sys
import numpy as np
from datetime import datetime
import re
import warnings
from shinier import __version__ as shinier_version
from shinier.Options import ACCEPTED_IMAGE_FORMATS, OPTION_TYPES
from shinier import ImageDataset, Options, ImageProcessor, REPO_ROOT
from shinier.utils import (
    Bcolors, console_log, load_np_array, colorize,
    print_shinier_header, generate_pydantic_key_value_dict
)

# Compute repo root as parent of /src/shinier/
IS_TTY = sys.stdin.isatty()
ACCEPTED_FORMATS = list(get_args(ACCEPTED_IMAGE_FORMATS))
ESCAPE_KEYS = ['q', 'exit']


#########################################
#            GENERIC PROMPT             #
#########################################
def prompt(
    label: str,
    default: Optional[Any] = None,
    kind: str = "str",
    choices: Optional[List[str]] = None,
    validator: Optional[Callable[[Any], Tuple[bool, str]]] = None,
    min_v: Optional[float] = None,
    max_v: Optional[float] = None,
    color: Optional[str] = None,
) -> Any:
    """Prompt user input with type casting, validation, and defaults.

    Args:
        label: Message displayed to the user.
        default: Default value returned if Enter is pressed.
        kind: Input type ('str', 'int', 'float', 'bool', 'choice').
        choices: List of valid string choices for 'choice' inputs.
        validator: Optional callable returning (ok, msg) for validation.
        min_v: Minimum numeric bound (for int/float).
        max_v: Maximum numeric bound (for int/float).
        color: Console text color.

    Returns:
        The validated and type-converted user input.
    """

    def print_answer(answer: str = '', prefix: str = 'Selected'):
        """
        Prints the given answer with a custom prefix and updates the console output.
        Args:
            answer (str): The string representing the answer to be displayed.
            prefix (str): A custom prefix to precede the answer.
        """
        tick = ''
        if IS_TTY:
            tick = '> '
            sys.stdout.write("\033[F")  # Move cursor up one line
            sys.stdout.write("\033[K")  # Clear the line

        console_log(f"{tick}{prefix}: {answer}")
        print('')

    # Display question
    default_str = colorize('default', Bcolors.DEFAULT_TEXT)
    console_log(f"{Bcolors.BOLD}{label} (Enter=[{Bcolors.DEFAULT_TEXT}{default}{Bcolors.ENDC}{Bcolors.BOLD}], q=quit):{Bcolors.ENDC}")

    # Check args and set choices
    if kind == 'choice' and choices is None:
        raise ValueError('Must provide `choices` when kind == "choice"')
    if kind == 'bool':
        accepted_choices = ['yes', 'no', 'y', 'n']
        if choices is not None:
            warnings.warn("choices are ignored for kind 'bool'")
        if not isinstance(default, str) or default.lower().strip() not in accepted_choices:
            raise ValueError('`default` should be a string ("y" or "n") for kind "bool"')

    if kind == "choice" and choices:
        for i, c in enumerate(choices, 1):
            new_default_str = f" [{default_str}]" if default == i else ""
            choice_color = Bcolors.DEFAULT_TEXT if default == i else Bcolors.CHOICE_VALUE
            choice_nb = colorize(str(i), choice_color)
            console_log(f"[{choice_nb}] {c}{new_default_str}", indent_level=0, color=Bcolors.BOLD)
    if kind == 'bool':
        choices = ['Yes', 'No']
        new_default = None
        for i, c in enumerate(choices):
            is_default = c[0].lower() in default.lower()
            new_default_str = f" [{default_str}]" if is_default else ""
            if new_default is None and is_default:
                new_default = i + 1
            choice_color = Bcolors.DEFAULT_TEXT if is_default else Bcolors.CHOICE_VALUE
            choice_str = f"{colorize(str(c[0].lower()), choice_color)}"
            console_log(f"[{choice_str}] {choices[i]}{new_default_str}", indent_level=0, color=Bcolors.BOLD)
        default = new_default

    if IS_TTY:
        print("> ", end="", flush=True)
    raw = input().strip()
    # raw = input("> ").strip()

    # ---- Exit ----
    if raw.lower().strip() in ESCAPE_KEYS:
        console_log("Exit requested (q).", color=Bcolors.FAIL)
        sys.exit(0)

    # ---- Default ----
    if raw == "":
        if default is not None:
            default_ = choices[default-1] if kind in ['bool', 'choice'] else default
            print_answer(default_, prefix="Default selected")
            return default
        console_log("Please enter a value or specify a default.", indent_level=1, color=Bcolors.FAIL)
        return prompt(label, default, kind, choices, validator, min_v, max_v, color)

    # ---- Type conversion ----
    try:
        if kind == "int":
            val = int(raw)
            if (min_v is not None and val < min_v) or (max_v is not None and val > max_v):
                raise ValueError
        elif kind == "float":
            val = float(raw)
            if (min_v is not None and val < min_v) or (max_v is not None and val > max_v):
                raise ValueError
        elif kind == "bool":
            if raw.lower() in ("y", "yes"):
                val = True
            elif raw.lower() in ("n", "no"):
                val = False
            else:
                raise ValueError
        elif kind == "choice":
            val = int(raw)
            if not (1 <= val <= len(choices)):
                raise ValueError
        elif kind == "tuple":
            tokens = re.findall(r"[-+]?\d*\.?\d+(?:[eE][-+]?\d+)?", raw)
            val = tuple(map(float, tokens))
            if len(val) == 0:
                raise ValueError("Expected a non-empty list of tuple. E.g.: `0, 2` or `(0, 2)` or `[0, 2]`")
        else:
            val = raw
    except ValueError:
        console_log("Invalid input.", indent_level=1, color=Bcolors.FAIL)
        return prompt(label, default, kind, choices, validator, min_v, max_v, color)

    # ---- Validation ----
    if validator:
        ok, msg = validator(val)
        if not ok:
            console_log(f"✗ {msg}", indent_level=1, color=Bcolors.FAIL)
            return prompt(label, default, kind, choices, validator, min_v, max_v, color)

    if kind in ['bool', 'choice']:
        prefix = 'Default selected' if val == default else 'Selected'
        print_answer(answer=choices[val-1], prefix=prefix)
    else:
        print_answer(answer=val, prefix="Answered")

    return val

def options_display(opts):
    """Display the options after images are processed.

    Args:
        opts (Options): Options object
    """
    types = ['io']
    if opts.whole_image != 1:
        types.append('mask')
    types += ['mode', 'color', 'dithering_memory']
    if opts.mode == 1:
        types.append('luminance')
    if opts.mode in (2, 5, 6, 7, 8):
        types.append('histogram')
    if opts.mode in (3, 4, 5, 6, 7, 8):
        types.append('fourier')
    types.append('misc')
    for option_type in types:
        console_log(f"\n---- {option_type}  ----", indent_level=1, color=Bcolors.OKBLUE)
        for key, value in dict(opts).items():
            if key in OPTION_TYPES[option_type]:
                if "target" in key:
                    if isinstance(value, np.ndarray):
                        value = "CUSTOM"
                    else:
                        value = "AVERAGE"
                console_log(f"{key:<20}: {value}", indent_level=1, color=Bcolors.OKBLUE)


#########################################
#            SHINIER CLI CORE           #
#########################################
def SHINIER_CLI(images: Optional[np.ndarray] = None, masks: Optional[np.ndarray] = None) -> ImageProcessor:
    """Interactive CLI to configure and run SHINIER processing.

    Args:
        images: Optional image array to bypass folder selection.
        masks: Optional mask array to bypass folder selection.

    Returns:
        Options: Configured SHINIER options object.
    """
    print_shinier_header(is_tty=IS_TTY, version=shinier_version)
    opts = Options()

    # --------- General I/O ---------
    if images is None:
        in_dir = prompt(f"Input folder (directory path)? Accepted formats: {ACCEPTED_FORMATS}", default=str(opts.input_folder), kind="str")
        opts.input_folder = Path(in_dir).expanduser().resolve()
        image_paths = get_image_list(opts.input_folder)
        color = Bcolors.OKGREEN if len(image_paths) > 0 else Bcolors.FAIL
        console_log(msg=f'\033[F-> {len(image_paths)} image(s) found in {opts.input_folder}\n', indent_level=0, color=color, strip=False)
    else:
        opts.input_folder = None

    out_dir = prompt("Output folder (directory path)?", default=str(opts.output_folder), kind="str")
    opts.output_folder = Path(out_dir).expanduser().resolve()

    # --------- Profile ---------
    prof = prompt("Options profile?", default=1, kind="choice", choices=["Default options", "Legacy options (will duplicate the Matlab SHINE TOOLBOX results)", "Customized options"])

    # --------- Mask ---------
    if prof != 1:
        whole = prompt("Binary ROI masks: Analysis run on selected pixels (e.g. pixels >= 127)", default=1, kind="choice", choices=[
            "No ROI mask: Whole images will be analyzed",
            "ROI masks: Analysis run on pixels != a pixel value you will provide",
            "ROI masks: Analysis run on pixels != most frequent pixel value in the image",
            "ROI masks: Masks loaded from the `MASK` folder and analysis run on pixels >= 127"
        ])

        if whole == 4:
            if masks is None:
                mdir = prompt(f"Masks folder (directory path)? Accepted formats: {ACCEPTED_FORMATS}. Will use ", default=str(REPO_ROOT / "data/MASK"), kind="str")
                opts.masks_folder = Path(mdir).expanduser().resolve()
                mask_paths = get_image_list(opts.masks_folder)
                color = Bcolors.OKGREEN if len(mask_paths) > 0 else Bcolors.FAIL
                console_log(msg=f'\033[F-> {len(mask_paths)} image(s) found in {opts.masks_folder}\n', indent_level=0, color=color, strip=False)
            else:
                opts.masks_folder = None
        opts.whole_image = [1, 2, 2, 3][whole-1]

        if whole in (2, 3):
            opts.background = opts.background if whole == 3 else prompt("ROI masks: Analysis will be run on pixels != [input a value between 0–255]", default=127, kind="int", min_v=0, max_v=255)

    # --------- Processing Mode ---------
    mode = prompt("Processing mode", default=2, kind="choice", choices=[
        "Luminance only (lum_match)",
        "Histogram only (hist_match)",
        "Spatial frequency only (sf_match)",
        "Spectrum only (spec_match)",
        "Histogram + Spatial frequency",
        "Histogram + Spectrum",
        "Spatial frequency + Histogram",
        "Spectrum + Histogram",
        "Dithering only"
    ])
    opts.mode = mode

    # --------- Legacy Mode ---------
    if prof == 2:
        opts.legacy_mode = True

    # --------- Custom Profile ---------
    if prof == 3:
        as_gray = prompt("Load images as grayscale?", default="No", kind="bool")
        opts.as_gray = as_gray == 1
        linear_luminance = prompt("Are pixel values linearly related to luminance?", default=2, kind='choice', choices=[
            f"{Bcolors.CHOICE_VALUE}Yes [legacy mode]{Bcolors.ENDC}\n\t- No color-space conversion.\n\t- Assuming input images are linear to luminance.\n\t- All transformations will be applied independently to each channel which may produce out-of-gamut values",
            f"{Bcolors.DEFAULT_TEXT}No [default]{Bcolors.ENDC}:\n\t- Assumes input images are regular sRGB images, i.e. gamma-encoded.\n\t- Images will first be converted into CIE xyY color-space\n\t- All transformations will be applied on the luminance channel (Y) of the CIE xyY color space.\n\t- Images are then reconverted into sRGB using transformed luminance channel (Y) and original chromatic channels (x, y),\n\t- This mode should preserves color gamuts",
        ])
        opts.linear_luminance = linear_luminance == 1
        if not opts.linear_luminance:
            rec_standard = prompt("Specifies the Rec. color standard used for RGB ↔ XYZ conversion (default = 2)", default=2, kind='choice', choices=[
                f"Rec.601 (SDTV) [{Bcolors.CHOICE_VALUE}legacy mode]{Bcolors.ENDC}",
                "Rec.709 (HDTV)",
                "Rec.2020 (UHDTV, wide-gamut HDR)"
            ])
            opts.rec_standard = rec_standard

        opts.conserve_memory = prompt("Conserve memory (creates a temporary directory and keep only one image in RAM)?", default='y', kind="bool")

        # Dithering
        dith_choices = ["No dithering", "Noisy-bit dithering", "Floyd–Steinberg dithering"]
        if mode != 9:
            dith = prompt("Apply dithering before final uint8 cast?", default=2,
                          kind="choice", choices=dith_choices)
            opts.dithering = dith - 1
        else:
            dith = prompt("Which dithering is going to be applied?", default=1,
                          kind="choice", choices=dith_choices[1:])
            opts.dithering = dith

        # Seed
        now = datetime.now()
        opts.seed = prompt("Provide seed for pseudo-random number generator or use time-stamped default", default=int(now.timestamp()), kind="int")

        # ---- Mode-Specific Options ----
        if mode == 1:
            opts.safe_lum_match = prompt("Safe luminance matching (will ensure pixel values fall within [0, 255])?", default='n', kind="bool")
            opts.target_lum = prompt("Target luminance list (mean, std)", default="0, 0", kind="tuple")

        if mode in (2, 5, 6, 7, 8):
            ho = prompt("Histogram specification with SSIM optimization (see Avanaki, 2009)?", default='y', kind="bool")
            opts.hist_optim = ho != 2
            if ho == 2:
                opts.hist_iterations = prompt("How many SSIM iterations?", default=5, kind="int", min_v=1, max_v=1_000_000)
                opts.step_size = prompt("What is the SSIM step size?", default=34, kind="int", min_v=1, max_v=1_000_000)
            opts.hist_specification = None
            if not opts.hist_optim:
                hs = prompt("Which histogram specification?", default=4, kind="choice", choices=[
                    "Exact with noise (legacy)",
                    "Coltuc with moving-average filters",
                    "Coltuc with gaussian filters",
                    "Coltuc with gaussian filters and noise if residual isoluminant pixels"
                ])
                opts.hist_specification = hs - 1

            thp1 = prompt("What should be the target histogram?", default=1, kind="choice", choices=[
                'Average histogram of input images',
                'Flat histogram a.k.a. `histogram equalization`',
                'Custom: You provide one as a .npy file'
            ])
            th = [None, 'equal'][thp1-1] if thp1 in [1, 2] else 'custom'
            if th == 'custom':
                thp2 = prompt("Path to target histogram (.npy/.txt/.csv)?", kind="str")
                th = load_np_array(thp2)
            if th is not None:
                opts.target_hist = th

        if mode in (3, 4, 5, 6, 7, 8):
            rsel = prompt("What type of rescaling after sf/spec?", default=2, kind="choice",
                          choices=["none", "min/max of all images", "avg min/max"])
            opts.rescaling = rsel - 1
            ans = prompt("Use a specific target spectrum?", default='n', kind="bool")
            if ans == 1:
                tsp = prompt("Path to target spectrum (.npy/.txt/.csv)?", kind="str")
                ts = load_np_array(tsp)
                if ts is not None:
                    opts.target_spectrum = ts

        if mode in (5, 6, 7, 8):
            opts.iterations = prompt("How many composite iterations (hist/spec coupling)?", default=2, kind="int", min_v=1, max_v=1_000_000)

    # ---- Progress info ----
    if prof != 1:
        prog_info = prompt('Select verbosity level', kind="choice", default=2, choices=[
                "None (quiet mode)",
                "Progress bar with ETA",
                "Basic progress steps (no progress bar)",
                "Detailed step-by-step info (no progress bar)",
                "Debug mode for developers (no progress bar)"
        ])
        opts.verbose = prog_info - 2

    # ---- Start SHINIER ----
    dataset = ImageDataset(images=images, masks=masks, options=opts) if (images or masks) else ImageDataset(options=opts)
    results = ImageProcessor(dataset=dataset, verbose=opts.verbose, from_cli=True)

    console_log("╔══════════════════════════════════════════════════════╗")
    console_log("║                      OPTIONS                         ║")
    console_log("╚══════════════════════════════════════════════════════╝")

    options_display(opts)

    return results


def get_image_list(src_path: Path):

    def is_image(file: Path):
        return file.is_file() and len(file.suffix) > 1 and file.suffix.lower()[1:] in ACCEPTED_FORMATS

    # Convert to Path if input_data is a string
    input_path = Path(src_path) if isinstance(src_path, str) else src_path

    # Handle cases with wildcards
    if "*" in str(input_path):  # Check if it's a glob pattern
        directory = input_path.parent
        pattern = input_path.name
        all_files = sorted(directory.glob(pattern))
    else:
        all_files = [input_path] if input_path.is_file() else sorted(input_path.glob("*"))

    # Filter to include only recognized image files
    return sorted([p for p in all_files if is_image(p)])


def main():
    import argparse
    parser = argparse.ArgumentParser(description="SHINIER")
    parser.add_argument("--show_results", action="store_true",
                        help="Display (or save) a summary figure showing before/after processing.")
    parser.add_argument("--save_path", type=str, default=None,
                        help="Save the overview figure instead of displaying it. Provide output path, e.g. figure.png")
    parser.add_argument("--image_index", type=int, default=0,
                        help="Image index for the overview figure (default: 0)")

    args = parser.parse_args()

    # Run the interactive core
    processor = SHINIER_CLI()

    # Overview mode
    if args.show_results:
        from shinier.utils import show_processing_overview
        try :
            import matplotlib.pyplot as plt
        except ImportError:
            raise RuntimeError(
                "Matplotlib is not installed. "
                "Install with: pip install shinier[viz]"
            )
        fig = show_processing_overview(processor, img_idx=args.image_index, show_figure=False)

        if args.save_path is not None:
            fig.savefig(args.save_path, dpi=150)
            print(f"Processing overview figure saved successfully at: {args.save_path}")
        else:
            plt.show()


if __name__ == "__main__":
    main()



"""
Color-space conversion tools for the SHINIER package.
"""

from .Converter import ColorConverter, ColorTreatment, COLOR_STANDARDS, WHITE_D65, M_RGB2XYZ_709, M_RGB2XYZ_2020, \
    M_RGB2XYZ_601, rgb2gray, gray2rgb, RGB_STANDARD, REC_STANDARD, RGB2GRAY_WEIGHTS

__all__ = [
    "ColorConverter",
    "ColorTreatment",
    "rgb2gray",
    "gray2rgb",
    "COLOR_STANDARDS",
    "WHITE_D65",
    'M_RGB2XYZ_601',
    "M_RGB2XYZ_2020",
    "M_RGB2XYZ_709",
    "REC_STANDARD",
    "RGB_STANDARD",
    "RGB2GRAY_WEIGHTS"
]


"""
SHINIER: Spectrum, Histogram, and Intensity Normalization, Equalization, and Refinement.

This package provides advanced image-processing utilities for luminance,
histogram, and spatial frequency normalization, adapted from the original
MATLAB SHINE Toolbox.

References:
    Willenbockel, V., Sadr, J., Fiset, D., Horne, G. O., Gosselin, F., & Tanaka, J. W. (2010).
    Controlling low-level image properties: The SHINE toolbox.
    *Behavior Research Methods, 42*(3), 671–684. https://doi.org/10.3758/BRM.42.3.671

    See accompanying paper: Salvas-Hébert, M., Dupuis-Roy, N., Landry, C., Charest, I. & Gosselin, F. (2025)
    The SHINIER the Better: An Adaptation of the SHINE Toolbox on Python.
"""

# Metadata
__author__ = "Nicolas Dupuis-Roy"
__version__ = "0.1.6"
__email__ = "nicolas.dupuis.roy@umontreal.ca"

# For direct importation
from importlib import util
from pathlib import Path
_HAS_CYTHON = util.find_spec("shinier._cconvolve") is not None

# This is the *package* root: src/shinier in dev, site-packages/shinier when installed
DEV_ROOT = Path(__file__).resolve().parents[2]
REPO_ROOT = Path(__file__).resolve().parent

if _HAS_CYTHON:
    from ._cconvolve import convolve2d_direct, convolve2d_separable
else:
    convolve2d_direct = None
    convolve2d_separable = None

__all__ = [
    "Options",
    "ImageDataset",
    "ImageListIO",
    "ImageProcessor",
    "convolve2d_direct",
    "convolve2d_separable",
    "_HAS_CYTHON",
    "color",
    "SHINIER_CLI",
    "REPO_ROOT",
    "DEV_ROOT",
    "__version__",
]

from .Options import Options
from .ImageDataset import ImageDataset
from .ImageListIO import ImageListIO
from .ImageProcessor import ImageProcessor
from .SHINIER import SHINIER_CLI
from . import color
